import pytest

from pyradios.utils import bool_to_string


@pytest.mark.parametrize(
    "inp, expected",
    [(True, "true"), (False, "false")]
)
def test_bool_to_string(inp, expected):
    assert bool_to_string(inp) == expected


@pytest.mark.parametrize(
    "inp, expected",
    [
        ("a", pytest.raises(TypeError)),
        (1, pytest.raises(TypeError)),
    ],
)
def test_bool_to_string_value_error(inp, expected):
    with expected as exc_info:
        bool_to_string(inp)
    assert "Value must be True or False." == str(exc_info.value)



import pytest
import sys
import logging
import random
import os
from functools import reduce


from pyradios import RadioBrowser, RadioFacets


log = logging.getLogger("pyradios")


@pytest.fixture
def rb():
    _rb = RadioBrowser()
    return _rb


def partition(predicate, iterable):
    """
    splits the items in iter into two partitions based on predicate() outcome
      True  --> into first list of returned tuple
      False --> into second list
    """
    def split(parts, item):
        parts[int(not predicate(item))].append(item)
        return parts
    return reduce(split, iterable, ([], []))


def test_facet_init(rb):
    log.info("test facet_init started")
    rf = RadioFacets(rb)
    assert rf.result is not None, "expecting to have a result-set"
    assert len(rf) > 0, "expecting the no-query result-set to not be empty"

    # split the set in stations with and without tags
    parts = partition(lambda s: len(s['tags']) == 0, rf.result)

    for part in parts:
        anystation = random.choice(part)
        log.debug(f"one rsult == {anystation}")

        assert rf.tags is not None, "expecting a tags histogram"
        anytag = random.choice(anystation['tags'].split(','))
        foundtags = list(filter(lambda t: t['name'] == anytag, rf.tags))
        assert len(foundtags) == 1, f"tag '{anytag}' not in the result-set"
        assert foundtags[0]['count'] > 1, f"not even one match '{anytag}'"

    assert rf.countrycodes is not None, "expecting a contrycode histogram"
    assert rf.languages is not None, "expecting a language histogram"
    assert rf.codecs is not None, "expecting a codec histogram"
    assert rf.states is not None, "expecting a state histogram"

    log.debug(f"facet repr == {rf}")
    log.debug(f"facet top tags == {rf.tags[:10]}")
    log.debug(f"facet top langs == {rf.languages[:10]}")
    log.debug(f"facet top codecs == {rf.codecs[:10]}")
    log.debug(f"facet top ctrys == {rf.countrycodes[:10]}")


def test_facet_narrow_broaden(rb):
    log.info("test facet_narrow_broaden started")
    limit = 10                         # not make this smaller then 5
    qry_be = dict(countrycode="be")    # the center of the world
    qry_nl = dict(language="dutch")    # spoken in .be, as is french and german
    qry_lim = dict(limit=limit)        # size limit
    qry_klara = dict(name="klara", name_exact=False)  # radiostation in be
    qry_clss = dict(tag="classical")   # tag that will fit radio klara

    # start off with a set of belgian stations
    rfbe = RadioFacets(rb, **qry_be)
    assert len(rfbe) > limit, f"execpting > {limit} stations in .be"
    log.debug(f"found {len(rfbe)} stations matching {rfbe.filter}")

    # narrow down to dutch speaking stations
    rfbenl = rfbe.narrow(**qry_nl)
    assert len(rfbenl) > limit, f"execpting > {limit} nl stations in .be"
    assert len(rfbenl) < len(rfbe), f"expecting set to be < {len(rfbe)}"
    log.debug(f"found {len(rfbenl)} stations matching {rfbenl.filter}")

    # add the size limit
    rfbenllim = rfbenl.narrow(**qry_lim)
    assert len(rfbenllim) == limit, f"expecting set to be size {limit}"
    log.debug(f"found {len(rfbenllim)} stations matching {rfbenllim.filter}")

    # further narrow to name-matches
    rfbenlklaralim = rfbenllim.narrow(**qry_klara)
    assert 0 < len(rfbenlklaralim) < limit, f"expecting 0 < size < {limit}"
    log.debug(f"found {len(rfbenlklaralim)} stations " +
              f"matching {rfbenlklaralim.filter}")

    # broaden up by skipping the country and language constraint
    rfklaralim = rfbenlklaralim.broaden(**qry_be, **qry_nl)
    assert len(rfbenlklaralim) <= len(rfklaralim) < limit
    log.debug(f"found {len(rfklaralim)} stations " +
              f"matching {rfklaralim.filter}")

    # exchange the limit constraint with a tag constraint
    rfklaraclss = rfklaralim.broaden(tuple(qry_lim.keys())).narrow(**qry_clss)
    assert len(rfklaraclss) <= len(rfklaralim)
    log.debug(f"found {len(rfklaraclss)} stations " +
              f"matching {rfklaraclss.filter}")


def enable_stdout_logging():
    if 'PYTEST_LOG' in os.environ:
        loglevel = logging.getLevelName(os.environ['PYTEST_LOG'].upper())
        log.setLevel(loglevel)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(loglevel)
        msg_fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = logging.Formatter(msg_fmt)
        handler.setFormatter(formatter)
        log.addHandler(handler)


if __name__ == "__main__":
    enable_stdout_logging()
    log.info(
        "Running tests in ",
        __file__,
        "with -v(erbose) and -s(no stdout capturing) ",
        "and logging to stdout, level controlled by env var ${PYTEST_LOG}")
    sys.exit(pytest.main(["-v", "-s",  __file__]))



import socket
from unittest.mock import Mock
from unittest.mock import patch

import pytest
from pyradios.base_url import fetch_all_hosts


def getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
    """
    returns:
        (family, type, proto, canonname, sockaddr)
    """
    return_value = [
        ("family", "type", "proto", "canonname", ("192.168.1.1", 80)),
        ("family", "type", "proto", "canonname", ("192.168.1.2", 80)),
        ("family", "type", "proto", "canonname", ("192.168.1.3", 80)),
        ("family", "type", "proto", "canonname", ("192.168.1.4", 80)),
    ]
    return return_value


def test_fetch_all_hosts():

    with patch("socket.getaddrinfo", getaddrinfo):
        assert fetch_all_hosts() == [
            "192.168.1.1",
            "192.168.1.2",
            "192.168.1.3",
            "192.168.1.4",
        ]


def test_fetch_all_hosts_network_failure():
    mock_getaddrinfo = Mock()
    mock_getaddrinfo.side_effect = socket.gaierror
    with patch("socket.getaddrinfo", mock_getaddrinfo):
        with pytest.raises(socket.gaierror) as exc_info:
            fetch_all_hosts()
        assert issubclass(exc_info.type, (OSError,))
        assert exc_info.type == socket.gaierror



from pyradios.radios import RadioBrowser
from itertools import chain
from functools import reduce
from collections import defaultdict
import logging


log = logging.getLogger("pyradios")


class RadioFacets:
    FACETS = ["tags", "countrycode", "language", "state", "codec"]

    def __init__(self, rb, **params):
        assert rb is not None, "facets requires a RadioBrowser service to call"
        assert isinstance(rb, RadioBrowser), "RadioBrowser service wrong type"
        self.rb = rb
        self.filter = params if params is not None else dict()
        self._fetch_and_calc()

    def __repr__(self):
        typename = type(self).__name__
        args = ['{}={!r}'.format(k, v) for k, v in self.filter.items()]
        return "%s(%s)" % (typename, ', '.join(args))

    def _fetch_and_calc(self):
        self.result = self.rb.search(**self.filter)
        self.facets = dict()

        def facetcount(facets, item):
            for f in facets.keys():
                for fv in item[f].split(','):
                    facets[f][fv] += 1  # tracks empty facet-values as ''
            return facets

        init = {f: defaultdict(int) for f in RadioFacets.FACETS}
        captured_hist_per_facet = reduce(facetcount, self.result, init)
        for f in captured_hist_per_facet.keys():
            # unsorted facet histogram
            usf_hist = captured_hist_per_facet[f].items()
            # order histogram items from high to low
            sf_hist = sorted(usf_hist, key=lambda i: i[1], reverse=True)
            # pluralize the key if needed
            fpub = f if f[-1] == 's' else f + "s"
            self.facets[fpub] = [{"name": k, "count": v} for k, v in sf_hist]

    def __getattr__(self, key):
        if key in self.facets:
            return self.facets[key]
        # else
        return getattr(self, key)

    def __len__(self):
        return len(self.result)

    def _derived(self, **filter):
        return RadioFacets(self.rb, **filter)

    def narrow(self, **pars2add):
        log.debug("flatten ++ params(%s)" % pars2add)
        filter = dict(self.filter)  # clone the dict
        filter.update(pars2add)
        return self._derived(**filter)

    def broaden(self, *keys2rm, **pars2rm):
        log.debug("broaden -- keys(%s) and params(%s)" % (keys2rm, pars2rm))
        filter = dict(self.filter)
        for k in chain(keys2rm, pars2rm.keys()):
            filter.pop(k, None)
        return self._derived(**filter)



import logging
from logging import NullHandler
from pyradios.radios import RadioBrowser
from pyradios.facets import RadioFacets

# warning: setup.py assumes strict single quotes on the line below
__version__ = '1.0.0'

__all__ = ["RadioBrowser", "RadioFacets"]

logging.getLogger(__name__).addHandler(NullHandler())



import re

import pytest
import random
import responses

from pyradios import RadioBrowser
from pyradios.radios import version


BASE_URL = "https://nl1.api.radio-browser.info/"


def pick_random_station(rb, **params):
    stations = rb.search(**params)
    return random.choice(stations)


def test_version():
    assert version == "1.0.0"


@pytest.fixture
def rb():
    _rb = RadioBrowser()
    _rb.base_url = BASE_URL
    return _rb


@pytest.fixture
def station_struct():
    """https://de1.api.radio-browser.info/#Struct_station"""
    return {
        "changeuuid",
        "stationuuid",
        "name",
        "url",
        "url_resolved",
        "homepage",
        "favicon",
        "tags",
        "country",
        "countrycode",
        "state",
        "language",
        "languagecodes",
        "votes",
        "lastchangetime",
        "lastchangetime_iso8601",
        "codec",
        "bitrate",
        "hls",
        "lastcheckok",
        "lastchecktime",
        "lastchecktime_iso8601",
        "lastcheckoktime",
        "lastcheckoktime_iso8601",
        "lastlocalchecktime",
        "lastlocalchecktime_iso8601",
        "clicktimestamp",
        "clicktimestamp_iso8601",
        "clickcount",
        "clicktrend",
        "ssl_error",
        "geo_lat",
        "geo_long",
        "serveruuid",
        "has_extended_info",
        "iso_3166_2",
    }


def test_station_struct(rb, station_struct):
    """
    Checks if the keys in the API response structure have changed.

    Note: This test is more about the API service implmentation, than
    it is about this python library.

    See details:
        https://de1.api.radio-browser.info/#Struct_station
    """
    stationuuid = "d97b5842-8e9b-46cc-85f0-d2dff6738c7c"
    resp = rb.station_by_uuid(stationuuid)
    response_keys = set(resp[0].keys())
    missing_keys = response_keys.difference(station_struct)
    assert len(missing_keys) == 0, f"Keys missing {missing_keys}"


@responses.activate
def test_request_station_click_counter(rb):

    expected = {
        "ok": True,
        "message": "retrieved station url",
        "stationuuid": "96062a7b-0601-11e8-ae97-52543be04c81",
        "name": "BBC Radio 1",
        "url": "http://bbcmedia.ic.llnwd.net/stream/bbcmedia_radio1_mf_p",
    }
    # mock responses
    responses.add(
        responses.GET,
        BASE_URL + "json/url/96062a7b-0601-11e8-ae97-52543be04c81",
        json=expected,
        status=200,
    )

    resp = rb.click_counter("96062a7b-0601-11e8-ae97-52543be04c81")

    assert "ok" in resp
    assert "message" in resp
    assert "stationuuid" in resp
    assert "name" in resp
    assert "url" in resp
    assert resp == expected


@responses.activate
def test_request_countrycodes(rb):

    expected = [{"name": "AD", "stationcount": 5}]
    responses.add(
        responses.GET,
        BASE_URL + "json/countrycodes/",
        json=expected,
        status=200,
    )

    resp = rb.countrycodes()
    assert len(resp) > 0, "at least one country should be in the response"
    assert "name" in resp[0]
    assert "stationcount" in resp[0]
    assert resp == expected


def test_request_countrycodes_with_filters(rb):

    # expected = [{"name": "BR", "stationcount": 607}]

    resp = rb.countrycodes(code="BR")
    assert len(resp) > 0, "at least one country should be in the response"
    assert resp[0]["name"] == "BR"


def test_request_codecs(rb):

    # expected = [{"name": "MP3", "stationcount": 16620}]

    resp = rb.codecs()
    assert len(resp) > 0, "at least one codec should be in the response"
    assert "name" in resp[0]
    assert "stationcount" in resp[0]


@responses.activate
def test_request_codecs_with_filters(rb):
    """
    This test runs against a subset of the API response, with the sole
    purpose of ensuring the behavior of the filter mechanism of the
    tested method.
    """
    # subset of the API response for testing purposes
    payload = [
        {'name': 'AAC', 'stationcount': 2226},
        {'name': 'MP3', 'stationcount': 19678},
        {'name': 'MP3,H.264', 'stationcount': 6},
        {'name': 'OGG', 'stationcount': 265},
        {'name': 'UNKNOWN', 'stationcount': 1827},

    ]
    responses.add(
        responses.GET,
        BASE_URL + "json/codecs/",
        json=payload,
        status=200,
    )

    resp = rb.codecs(codec="mp3")
    assert len(resp) == 1
    assert resp[0]["name"] == "MP3", "only one codec should be in the response"


def test_request_states_with_filters(rb):

    # expected = [{"name": "Parana", "country": "Brazil", "stationcount": 23}]

    resp = rb.states(country="BRAZIL", state="Parana")

    assert len(resp) > 0, "at least one state should be in the response"
    assert "name" in resp[0]
    assert "country" in resp[0]
    assert "stationcount" in resp[0]


def test_request_languages_with_filters(rb):

    # expected = [
    #     {"name": "brazilian portuguese", "stationcount": 8,},
    #     {"name": "portuguese", "stationcount": 638,},
    # ]

    resp = rb.languages(language="portuguese")

    assert len(resp) > 0, "at least one language should be in the response"
    assert "name" in resp[0]
    assert "stationcount" in resp[0]


# @responses.activate
def test_request_tags_with_filters(rb):

    # expected = [{"name": "drum and bass", "stationcount": 73,}]
    # responses.add(
    #     responses.GET,
    #     BASE_URL + "json/tags/drum%20and%20bass",
    #     json=expected,
    #     status=200,
    # )
    resp = rb.tags(tag="Drum and Bass")

    assert len(resp) > 0, "at least one tag should be in the response"
    assert "name" in resp[0]
    assert "stationcount" in resp[0]


def test_request_stations_by_votes(rb):
    resp = rb.stations_by_votes(1)
    assert len(resp) == 1, "exactly one station should be in the response"


def test_request_station_byuuid(rb):
    uuid = pick_random_station(rb, limit=100)["stationuuid"]
    resp = rb.station_by_uuid(uuid)
    assert len(resp) == 1, "exactly one should match the uuid '%s'" % uuid


def test_request_station_tag_list(rb):
    tag_list = ["New York City", "Jazz"]
    resp = rb.stations_by_tag_list(tag_list)

    assert len(resp) > 0, "no match on tags [%s]" % ", ".join(tag_list)

    some_stations = random.choices(resp, k=5)  # just take some stations

    for station in some_stations:
        # Be aware that tag search is token (or word) based
        # meaning that searching 'jazz' will match 'smooth jazz'
        tagtokens = re.split(r"\s|,", station["tags"])
        queried_tokens = re.split(r"\s|,", ",".join(tag_list).lower())
        missing_tokens = set(queried_tokens) - set(tagtokens)
        assert (
            len(missing_tokens) == 0
        ), "tokens missing %s in result-tag %s" % (missing_tokens, tagtokens)



import io
import re
from setuptools import setup
from setuptools import find_packages


DESCRIPTION = "Python client for the Radio Browser API"


def get_version():
    content = open("pyradios/__init__.py").read()
    mo = re.search(r"__version__\s+=\s+'([^']+)'", content)
    if not mo:
        raise RuntimeError(
                'Unable to find version string in pyradios/__init__.py'
            )
    return mo[1]


def readme():
    with io.open("README.md", "r", encoding="utf-8") as f:
        return f.read()


def required(sfx=''):
    with open(f"requirements{sfx}.txt") as f:
        return f.read().splitlines()


setup(
    name="pyradios",
    version=get_version(),
    description=DESCRIPTION,
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="pyradios wrapper radios api",
    author="André P. Santos",
    author_email="andreztz@gmail.com",
    url="https://github.com/andreztz/pyradios",
    license="MIT",
    packages=find_packages(),
    install_requires=required(),
    extras_require={'dev': required('-dev')},
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    python_requires=">=3.6",
    project_urls={
        "Source": "https://github.com/andreztz/pyradios/",
        "Upstream": "https://api.radio-browser.info/",
    },
)
